# -*- coding=utf-8 -*-

"""Un-redondant contig graph module."""


from pathlib import Path
from typing import Iterator, Literal

from revsymg.graphs import RevSymGraph
from revsymg.index_lib import FORWARD_INT, ORIENT_REV, IndexT, OrT

from khloraascaf.inputs import (
    MULT_UPB_DEF,
    PRESSCORE_UPB_DEF,
    IdCT,
    MultT,
    PresScoreT,
    read_contig_attributes,
    read_contig_links_file,
)


# DOCU missing constant docstrings
# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Indexed                                   #
# ---------------------------------------------------------------------------- #
OccOrCT = tuple[IndexT, OrT, IndexT]
EOccOrCT = tuple[OccOrCT, OccOrCT]


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Indices                                   #
# ---------------------------------------------------------------------------- #
#
# For OccOrCT type
#
CIND_IND: Literal[0] = 0
COR_IND: Literal[1] = 1
COCC_IND: Literal[2] = 2

# ---------------------------------------------------------------------------- #
#                                  Attributes                                  #
# ---------------------------------------------------------------------------- #
MULT_ATTR = 'multiplicity'
PRESSCORE_ATTR = 'presence_score'
CTG_ID_ATTR = 'contig_id'
LINK_ID_ATTR = 'link_id'


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                        Multiplied Doubled Contig Graph                       #
# ---------------------------------------------------------------------------- #
class MDCGraph(RevSymGraph):
    """MDCGraph object.

    DOCU update MDCGraph's doc
    """

    def __init__(self, mult_upb: MultT = MULT_UPB_DEF,
                 presscore_upb: PresScoreT = PRESSCORE_UPB_DEF):
        super().__init__()
        #
        # List of v_ind | v_mult > 1
        #
        self.__repeated_contigs: list[IndexT] = []
        #
        # New attributes
        #
        self._vertices.new_attr(MULT_ATTR, mult_upb)
        self._vertices.new_attr(PRESSCORE_ATTR, presscore_upb)

    # ~*~ Dynamics ~*~

    def add_multiplied_vertex(self, mult: MultT,
                              presence_score: PresScoreT) -> IndexT:
        """Add a vertex.

        Parameters
        ----------
        mult : MultT
            Contig's multiplicity
        presence_score : PresScoreT
            Contig's presence score

        Returns
        -------
        IndexT
            Contig's index
        """
        contig_ind = self._vertices.add()
        self._vertices.set_attr(contig_ind, MULT_ATTR, mult)
        self._vertices.set_attr(contig_ind, PRESSCORE_ATTR, presence_score)
        if mult > 1:
            self.__add_repeated_contig(contig_ind)
        return contig_ind

    # ~*~ Iterators ~*~

    def multiplied_vertices(self) -> Iterator[OccOrCT]:
        """Iterate over the multiplied oriented contigs.

        Yields
        ------
        OccOrCT
            Multiplied oriented contig
        """
        for v_ind, v_or in self._vertices:
            for occ in range(self._vertices.attr(v_ind, MULT_ATTR)):
                yield v_ind, v_or, occ

    def multiplied_edges(self) -> Iterator[EOccOrCT]:
        """Iterate on edges of OccOrCT type.

        Yields
        ------
        EOccOrCT
            Edge of multiplied oriented contigs
        """
        for v_ind, v_or in self._vertices:
            v_mult = self._vertices.attr(v_ind, MULT_ATTR)
            for (w_ind, w_or), _ in self._edges.succs((v_ind, v_or)):
                w_mult = self._vertices.attr(w_ind, MULT_ATTR)
                for v_occ in range(v_mult):
                    for w_occ in range(w_mult):
                        yield (
                            (v_ind, v_or, v_occ),
                            (w_ind, w_or, w_occ),
                        )

    def multiplied_preds(self, v: OccOrCT) -> Iterator[OccOrCT]:
        """Iterate on the OccOrCT predecessors of `v`.

        Parameters
        ----------
        v : OccOrCT
            Multiplied oriented contig

        Yield
        -----
        OccOrCT
            Multiplied oriented predecessors
        """
        for (u_ind, u_or), _ in self._edges.preds((v[CIND_IND], v[COR_IND])):
            u_mult = self._vertices.attr(u_ind, MULT_ATTR)
            for u_occ in range(u_mult):
                yield u_ind, u_or, u_occ

    def multiplied_succs(self, v: OccOrCT) -> Iterator[OccOrCT]:
        """Iterate on the OccOrCT successors of `v`.

        Parameters
        ----------
        v : OccOrCT
            Multiplied oriented contig

        Yield
        -----
        OccOrCT
            Multiplied oriented sucessors
        """
        for (w_ind, w_or), _ in self._edges.succs((v[CIND_IND], v[COR_IND])):
            w_mult = self._vertices.attr(w_ind, MULT_ATTR)
            for w_occ in range(w_mult):
                yield w_ind, w_or, w_occ

    # ~*~ Getter ~*~

    def repeated_contigs(self) -> list[IndexT]:
        """Return the list of repeated contigs' indices.

        Returns
        -------
        list of IndexT
            List of repeated contigs' indices

        Note
        ----
        The list is sorted
        """
        return self.__repeated_contigs

    def multiplied_card(self) -> int:
        """The number of multiplied oriented contig.

        Returns
        -------
        int
            The number of vertices in multiplied doubled contig graph
        """
        return 2 * sum(
            self._vertices.attr(v_ind, MULT_ATTR)
            for v_ind in range(self._vertices.card_index())
        )

    # ~*~ Private ~*~

    def __add_repeated_contig(self, v_ind: IndexT):
        """Add the contig index in the repeat list.

        Parameters
        ----------
        v_ind : IndexT
            Contig's index

        Note
        ----
        For all i < j, repeated_contigs[i] < repeated_contigs[j]
        """
        self.__repeated_contigs.append(v_ind)


# ---------------------------------------------------------------------------- #
#                 Identifier Container Associated With MDCGraph                #
# ---------------------------------------------------------------------------- #
class MDCGraphIDContainer():
    """Container for identifiers linked to MDCgraph object.

    For each vertex' and edge' index in the graph is respectively associated
    a contig' and a link' identifier.
    """

    __NULL_ID: IdCT = ''

    def __init__(self):
        """The Initializer."""
        #
        # Contigs
        #
        #   forward contig indices (there is no index for the reverse)
        self.__contig_to_vertex: dict[IdCT, IndexT] = {}
        #   identifiers associated with vertex's index
        self.__vertex_to_contig: list[IdCT] = []
        #
        # Links
        #
        #   forward link indices (there is no index for the reverse)
        self.__edge_to_link: list[IdCT] = []

    # ~*~ Setter ~*~

    def associate_contig_vertex(self, contig_id: IdCT, vertex_ind: IndexT):
        """Associate the contig with the vertex.

        Parameters
        ----------
        contig_id : IdCT
            Contig's identifier
        vertex_ind : IndexT
            Vertex' index
        """
        self.__contig_to_vertex[contig_id] = vertex_ind
        while vertex_ind > len(self.__vertex_to_contig) - 1:
            self.__vertex_to_contig.append(self.__NULL_ID)
        self.__vertex_to_contig[vertex_ind] = contig_id

    def associate_link_edge(self, link_id: IdCT, edge_ind: IndexT):
        """Associate the link with the edge.

        Parameters
        ----------
        link_id : IdCT
            Link's identifier
        edge_ind : IndexT
            Edge's index
        """
        while edge_ind > len(self.__edge_to_link) - 1:
            self.__edge_to_link.append(self.__NULL_ID)
        self.__edge_to_link[edge_ind] = link_id

    # ~*~ Getter ~*~

    def contig_to_vertex(self, contig_id: IdCT) -> IndexT:
        """Returns the vertex index associated with the contig's id.

        Parameters
        ----------
        contig_id : IdCT
            Contig's identifier

        Returns
        -------
        IndexT
            Vertex' index
        """
        return self.__contig_to_vertex[contig_id]

    def vertex_to_contig(self, vertex_ind: IndexT) -> IdCT:
        """Returns the contig's id associated with the vertex index.

        Parameters
        ----------
        vertex_ind : IndexT
            Vertex' index

        Returns
        -------
        IdCT
            Contig's identifier
        """
        return self.__vertex_to_contig[vertex_ind]

    def edge_to_link(self, edge_index: IndexT) -> IdCT:
        """Returns the link's id associated with the edge index.

        Parameters
        ----------
        edge_index : IndexT
            Edge's index

        Returns
        -------
        IdCT
            Link's identifier
        """
        return self.__edge_to_link[edge_index]


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                Type Operation                                #
# ---------------------------------------------------------------------------- #
def rev_occorc(v: OccOrCT) -> OccOrCT:
    """Return the reverse orientation of v.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig

    Returns
    -------
    OccOrCT
        Reverse of the multiplied oriented contig
    """
    return v[CIND_IND], ORIENT_REV[v[COR_IND]], v[COCC_IND]


def first_forward_occurrence(vertex_index: IndexT) -> OccOrCT:
    """Returns the first forward occurrence with given vertex index.

    Parameters
    ----------
    vertex_index : IndexT
        Contig's index

    Returns
    -------
    OccOrCT
        The first occurrence, in forward orientation with the given vertex index
    """
    return vertex_index, FORWARD_INT, 0


# ---------------------------------------------------------------------------- #
#                                 Constructors                                 #
# ---------------------------------------------------------------------------- #
def mdcg_with_id_from_input_files(
    contig_attrs: Path,
    contig_links: Path,
    multiplicity_upperbound: MultT = MULT_UPB_DEF,
    presence_score_upperbound: PresScoreT = PRESSCORE_UPB_DEF) -> (
        tuple[MDCGraph, MDCGraphIDContainer]):
    """Instantiate a graph with identifiers from input files.

    Parameters
    ----------
    contig_attrs : Path
        Contigs' attributes file path
    contig_links : Path
        Contigs' links file path
    multiplicity_upperbound : MultT, optional
        Multiplicities upper bound, by default MULT_UPB_DEF
    presence_score_upperbound : PresScoreT, optional
        Multiplicities upper bound, by default MULT_UPB_DEF

    Returns
    -------
    MDCGraph
        Multiplied doubled contig graph
    MDCGraphIDContainer
        Identifiers container for the graph
    """
    mdcg = MDCGraph(
        mult_upb=multiplicity_upperbound,
        presscore_upb=presence_score_upperbound,
    )
    id_container = MDCGraphIDContainer()

    for contig_id, mult, presscore in read_contig_attributes(contig_attrs):
        id_container.associate_contig_vertex(
            contig_id,
            mdcg.add_multiplied_vertex(mult, presscore),
        )

    for l_id, c_id, c_or, d_id, d_or in read_contig_links_file(contig_links):
        id_container.associate_link_edge(
            l_id,
            mdcg.edges().add(
                (id_container.contig_to_vertex(c_id), c_or),
                (id_container.contig_to_vertex(d_id), d_or),
            ),
        )

    return mdcg, id_container



# -*- coding=utf-8 -*-

"""Set definitions for ILP invf model."""


from typing import Iterator

from revsymg.index_lib import FORWARD_INT, REVERSE_INT

from khloraascaf.inputs import MultT
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    COCC_IND,
    COR_IND,
    MULT_ATTR,
    EOccOrCT,
    IndexT,
    MDCGraph,
    OccOrCT,
)


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Sets                                     #
# ---------------------------------------------------------------------------- #
InvFT = tuple[OccOrCT, OccOrCT]
PInvFT = tuple[InvFT, InvFT]

# ---------------------------------------------------------------------------- #
#                                Forbidden Cases                               #
# ---------------------------------------------------------------------------- #
ForbiddenPairInvFT = tuple[IndexT, OccOrCT, OccOrCT, OccOrCT, OccOrCT]


# ============================================================================ #
#                            INVERTED FRAGMENTS SETS                           #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     InvF                                     #
# ---------------------------------------------------------------------------- #
def invf(mdcg: MDCGraph) -> Iterator[InvFT]:
    """Iterate on unredundant inverted fragments.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph

    Yields
    ------
    InvFT
        Inverted fragment
    """
    for v_ind in mdcg.repeated_contigs():
        yield from v_invf(mdcg, v_ind)


def v_invf(mdcg: MDCGraph, v_ind: IndexT) -> Iterator[InvFT]:
    """Iterate on inverted fragments of a vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    v_ind : IndexT
        Vertex index

    Yields
    ------
    InvFT
        Inverted fragment
    """
    v_mult = mdcg.vertices().attr(v_ind, MULT_ATTR)
    for i in range(0, 2 * (v_mult // 2), 2):
        yield (
            (v_ind, FORWARD_INT, i),
            (v_ind, REVERSE_INT, i + 1),
        )


def invf_builder(v: OccOrCT) -> tuple[InvFT, int]:
    """Inverted fragments builder from vertex.

    Parameters
    ----------
    v : OccOrCT
        Multiplied doubled contig

    Returns
    -------
    InvFT
        Inverted pair associated to vertex
    int
        Position of vertex in built pair

    Notes
    -----
    Requires that :math:`(v_{occ} - v_{or})|2`
    """
    return (
        ((v[CIND_IND], FORWARD_INT, v[COCC_IND] - v[COR_IND]),
         (v[CIND_IND], REVERSE_INT, v[COCC_IND] + (REVERSE_INT - v[COR_IND])),
         ),
        v[COR_IND],
    )


def invf_canonical(v: OccOrCT) -> OccOrCT:
    """Inverted fragment canonical.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig

    Returns
    -------
    OccOrCT
        Canonical of the inverted fragments

    Notes
    -----
    Requires that :math:`(v_{occ} - v_{or})|2`
    """
    return (v[CIND_IND], FORWARD_INT, v[COCC_IND] - v[COR_IND])


def invf_other(v: OccOrCT) -> OccOrCT:
    """Return the other vertex in the inverted fragments.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig

    Returns
    -------
    OccOrCT
        The other vertex in the inverted fragments

    Notes
    -----
    Requires that :math:`(v_{occ} - v_{or})|2`
    """
    if v[COR_IND] == REVERSE_INT:
        return (v[CIND_IND], FORWARD_INT, v[COCC_IND] - REVERSE_INT)
    return (v[CIND_IND], REVERSE_INT, v[COCC_IND] + REVERSE_INT)


# ---------------------------------------------------------------------------- #
#                                     PInvF                                    #
# ---------------------------------------------------------------------------- #
def pinvf(mdcg: MDCGraph) -> Iterator[PInvFT]:
    """Iterate on pair of inverted fragments.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph

    Yields
    ------
    PInvFT
        Pair of inverted fragments
    """
    repeated_contigs = mdcg.repeated_contigs()
    for i, v_ind in enumerate(repeated_contigs):
        for p in v_invf(mdcg, v_ind):
            # same vertex index
            for q in v_invf(mdcg, v_ind):
                if p[REVERSE_INT][COCC_IND] < q[FORWARD_INT][COCC_IND]:
                    yield p, q
            # superior vertex index
            for j in range(i + 1, len(repeated_contigs)):
                for q in v_invf(mdcg, repeated_contigs[j]):
                    yield p, q


# ---------------------------------------------------------------------------- #
#                                     AInvF                                    #
# ---------------------------------------------------------------------------- #
def ainvf(mdcg: MDCGraph) -> Iterator[EOccOrCT]:
    """Iterate on inverted fragments successors.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contigs graph

    Yields
    ------
    EOccOrCT
        Canonical edges between two inverted fragments
    """
    for u_ind in mdcg.repeated_contigs():
        u_mult = mdcg.vertices().attr(u_ind, MULT_ATTR)
        yield from __ainvf_succs_f(mdcg, u_ind, u_mult)
        yield from __ainvf_succs_r(mdcg, u_ind, u_mult)


def __ainvf_succs_f(mdcg: MDCGraph,
                    u_ind: IndexT, u_mult: MultT) -> Iterator[EOccOrCT]:
    """Iterate of successors of forward vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contigs graph
    u_ind : IndexT
        Vertex identifier
    u_mult : MultT
        Vertex multiplicity

    Yields
    ------
    EOccOrCT
        Canonical edges between two inverted fragments
    """
    # DOCU explain this
    for (v_ind, v_or), _ in mdcg.edges().succs((u_ind, FORWARD_INT)):
        if u_ind < v_ind:
            v_mult = mdcg.vertices().attr(v_ind, MULT_ATTR)
            for u_occ in range(0, 2 * (u_mult // 2), 2):
                for v_occ in range(0, 2 * (v_mult // 2), 2):
                    yield (
                        (u_ind, FORWARD_INT, u_occ),
                        (v_ind, v_or, v_occ + v_or),
                    )
        elif u_ind == v_ind:
            for u_occ in range(0, 2 * (u_mult // 2) - 2, 2):
                for v_occ in range(u_occ + 2, 2 * (u_mult // 2), 2):
                    yield (
                        (u_ind, FORWARD_INT, u_occ),
                        (u_ind, v_or, v_occ + v_or),
                    )


def __ainvf_succs_r(mdcg: MDCGraph,
                    u_ind: IndexT, u_mult: MultT) -> Iterator[EOccOrCT]:
    """Iterate of successors of reverse vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contigs graph
    u_ind : IndexT
        Vertex identifier
    u_mult : MultT
        Vertex multiplicity

    Yields
    ------
    EOccOrCT
        Canonical edges between two inverted fragments
    """
    # DOCU explain this
    for (w_ind, w_or), _ in mdcg.edges().succs((u_ind, REVERSE_INT)):
        if u_ind < w_ind:
            w_mult = mdcg.vertices().attr(w_ind, MULT_ATTR)
            for u_occ in range(0, 2 * (u_mult // 2), 2):
                for w_occ in range(0, 2 * (w_mult // 2), 2):
                    yield (
                        (u_ind, REVERSE_INT, u_occ + REVERSE_INT),
                        (w_ind, w_or, w_occ + w_or),
                    )
        elif u_ind == w_ind and w_or == FORWARD_INT:
            for u_occ in range(0, 2 * (u_mult // 2) - 2, 2):
                for w_occ in range(u_occ + 2, 2 * (u_mult // 2), 2):
                    yield (
                        (u_ind, REVERSE_INT, u_occ),
                        (u_ind, FORWARD_INT, w_occ),
                    )


def ainvf_other(edge: EOccOrCT) -> EOccOrCT:
    """Get the pairing edge.

    Parameters
    ----------
    edge : EOccOrCT
        Edge between two inverted fragments

    Returns
    -------
    EOccOrCT
        The other edge between the two inverted fragments
    """
    return invf_other(edge[1]), invf_other(edge[0])


# ---------------------------------------------------------------------------- #
#                               Consecutive PInvF                              #
# ---------------------------------------------------------------------------- #
def consecutive_pinvf(mdcg: MDCGraph, v_ind: IndexT) -> Iterator[PInvFT]:
    """Iterate over consecutive pairs of inverted fragments.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    v_ind : IndexT
        Vertex' index

    Yields
    ------
    PInvFT
        Consecutive pairs of inverted fragments
    """
    mult_v = mdcg.vertices().attr(v_ind, MULT_ATTR)
    for (i, j) in v_invf(mdcg, v_ind):
        if j[COCC_IND] + 2 < mult_v:
            yield (
                (i, j),
                (
                    (i[CIND_IND], i[COR_IND], i[COCC_IND] + 2),
                    (j[CIND_IND], j[COR_IND], j[COCC_IND] + 2),
                ),
            )


# ============================================================================ #
#                               FORBIDDEN PAIRING                              #
# ============================================================================ #
def pinvf_alpha(mdcg: MDCGraph) -> Iterator[EOccOrCT]:
    """Iterate on alpha keys for inverted fragments pairs.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph

    Yields
    ------
    EOccOrCT
        Alpha key set's element
    """
    for (i, j), (k, l) in pinvf(mdcg):
        yield i, k
        yield i, l
        yield j, k
        yield j, l


def forbidden_intersection(p_invf: PInvFT) -> Iterator[ForbiddenPairInvFT]:
    """Iterate over the forbidden cases.

    Parameters
    ----------
    p_invf : PInvFT
        Pair of inverted fragments

    Yields
    ------
    ForbiddenPairInvFT
        Forbidden pairing cases for inverted fragments,
        enriched with an index
    """
    (i, j), (k, l) = p_invf
    yield 0, i, k, j, l
    yield 1, i, l, j, k
    yield 2, j, k, i, l
    yield 3, j, l, i, k
    yield 4, k, i, l, j
    yield 5, k, j, l, i
    yield 6, l, i, k, j
    yield 7, l, j, k, i



# -*- coding=utf-8 -*-

"""Utilitary module for debugging."""


from pathlib import Path
from typing import Iterable, Iterator, Union

from revsymg.index_lib import IndexT

from khloraascaf.ilp.pulp_var_db import PuLPVarDirF, PuLPVarInvF
from khloraascaf.lib import RegionIDT
from khloraascaf.multiplied_doubled_contig_graph import OccOrCT
from khloraascaf.outputs import ORIENT_INT_STR, STR_ORIENT
from khloraascaf.result import ScaffoldingResult


# DOCU module debug
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
#
# Vertices of the regions
#
VERTICES_OF_REGIONS_PREFIX = 'vertices_of_regions'
"""Prefix of vertices of regions file name."""

VERTICES_OF_REGIONS_EXT = 'tsv'
"""Extension of vertices of regions file."""
#
# Repeated fragments
#
INVF_CODE = 'invf'
"""Inverted fragments code."""

DIRF_CODE = 'dirf'
"""Direct fragments code."""

FOUND_REPFRAG_PREFIX = 'repfrag'
"""Prefix of found repeated fragments file name."""

FOUND_REPFRAG_EXT = 'tsv'
"""Extension of found repeated fragments file."""


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                              Vertices Of Regions                             #
# ---------------------------------------------------------------------------- #
def write_vertices_of_regions(result: ScaffoldingResult,
                              vertices_of_regions_path: Path):
    """Write the regions' vertex file.

    #DOCU describe the output file here

    Parameters
    ----------
    result : ScaffoldingResult
        Previous scaffolding result
    vertices_of_regions_path : Path
        Path of the future file containing the regions' vertices
    """
    # TOTEST unit test write_vertices_of_regions
    with open(vertices_of_regions_path, 'w', encoding='utf-8') as f_out:
        for region_index in range(result.number_regions()):
            line: str = ''
            for v_ind, v_or, v_occ in result.region_occorc(region_index):
                line += f'{v_ind}\t{ORIENT_INT_STR[v_or]}\t{v_occ}\t'
            f_out.write(line[:-1] + '\n')


def read_vertices_of_regions(vertices_of_regions_path: Path) -> (
        Iterator[list[OccOrCT]]):
    """Read the regions' vertices file.

    Parameters
    ----------
    vertices_of_regions_path : Path
        Path of the file containing the regions' vertices

    Yields
    ------
    list of OccOrCT
        List of vertices of one region
    """
    with open(vertices_of_regions_path, 'r', encoding='utf-8') as f_in:
        for line in f_in:
            vertices_of_region: list[OccOrCT] = []
            l_occorcstr = line.split('\t')
            k = 0
            while k < len(l_occorcstr) - 2:
                vertices_of_region.append(
                    (
                        IndexT(l_occorcstr[k]),
                        STR_ORIENT[l_occorcstr[k + 1]],  # type: ignore
                        IndexT(l_occorcstr[k + 2]),
                    ),
                )
                k += 3
            yield vertices_of_region


def fmt_vertices_of_regions_filename(
        instance_name: str,
        ilp_combination: Iterable[RegionIDT]) -> str:
    """Format the vertices of regions file name.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of RegionIDT
        ILP string code combination

    Returns
    -------
    str
        Formatted file name
    """
    # TOTEST unit test fmt_vertices_of_regions_filename
    return (
        f'{VERTICES_OF_REGIONS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{VERTICES_OF_REGIONS_EXT}'
    )


# ---------------------------------------------------------------------------- #
#                           Found Repeated Fragments                           #
# ---------------------------------------------------------------------------- #
def write_found_repeated_fragments(var: Union[PuLPVarInvF, PuLPVarDirF],
                                   repeated_fragments_path: Path):
    r"""Write the repeated fragments found when solving a problem.

    DOCU file format
    * Format:
        ```md
        <code>\t<canonical>\n
        ```
    * Example:
        ```
        invf    1   +   2

        ```

    Parameters
    ----------
    var : PuLPVarInvF | PuLPVarDirF
        PuLP variable
    repeated_fragments_path : Path
        Path of the future file containing the repeated fragments solution
    """
    # TOTEST unit test write_found_repeated_fragments
    with open(repeated_fragments_path, 'w', encoding='utf-8') as f_out:
        if isinstance(var, PuLPVarInvF):
            for (v_ind, v_or, v_occ), _ in var.invf_solution():
                f_out.write(
                    f'{INVF_CODE}'
                    f'\t{v_ind}\t{ORIENT_INT_STR[v_or]}\t{v_occ}\n',
                )
        else:
            for (v_ind, v_or, v_occ), _ in var.dirf_solution():
                f_out.write(
                    f'{DIRF_CODE}'
                    f'\t{v_ind}\t{ORIENT_INT_STR[v_or]}\t{v_occ}\n',
                )


def read_found_repeated_fragments(
        repeated_fragments_path: Path) -> Iterator[tuple[str, OccOrCT]]:
    """Iterate over repeated fragments code and canonical.

    Parameters
    ----------
    repeated_fragments_path : Path
        Path of the file containing the repeated fragments solution

    Yields
    ------
    str
        Repeated fragments code
    OccOrCT
        Repeat fragments canonical
    """
    # TOTEST unit test read_found_repeated_fragments
    with open(repeated_fragments_path, 'r', encoding='utf-8') as f_in:
        for line in f_in:
            code_canonical = line.split()
            yield code_canonical[0], (
                IndexT(code_canonical[1]),
                STR_ORIENT[code_canonical[2]],  # type: ignore
                IndexT(code_canonical[3]),
            )


def fmt_found_repeated_fragments_filename(
        instance_name: str,
        ilp_combination: Iterable[RegionIDT]) -> str:
    """Format the found repeated fragments file name.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of RegionIDT
        ILP string code combination

    Returns
    -------
    str
        Formatted file name
    """
    # TOTEST unit test fmt_found_repeated_fragments_filename
    return (
        f'{FOUND_REPFRAG_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{FOUND_REPFRAG_EXT}'
    )



# -*- coding=utf-8 -*-

"""Set definitions for ILP invf model."""


from typing import Iterator

from revsymg.index_lib import FORWARD_INT, REVERSE_INT

from khloraascaf.inputs import MultT
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    COCC_IND,
    COR_IND,
    MULT_ATTR,
    EOccOrCT,
    IndexT,
    MDCGraph,
    OccOrCT,
)


# ============================================================================ #
#                                     TYPES                                    #
# ============================================================================ #
DirFT = tuple[OccOrCT, OccOrCT]
PDirFT = tuple[DirFT, DirFT]

# ---------------------------------------------------------------------------- #
#                                Forbidden Cases                               #
# ---------------------------------------------------------------------------- #
ForbiddenPairDirFT = tuple[IndexT, OccOrCT, OccOrCT, OccOrCT, OccOrCT]


# ============================================================================ #
#                             DIRECT FRAGMENTS SETS                            #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     DirF                                     #
# ---------------------------------------------------------------------------- #
def dirf(mdcg: MDCGraph) -> Iterator[DirFT]:
    """Iterate on unredundant direct fragments.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph

    Yields
    ------
    DirFT
        Direct fragment
    """
    for v_ind in mdcg.repeated_contigs():
        yield from v_dirf(mdcg, v_ind)


def v_dirf(mdcg: MDCGraph, v_ind: IndexT) -> Iterator[DirFT]:
    """Iterate on direct fragments of a vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    v_ind : IndexT
        Vertex index

    Yields
    ------
    DirFT
        Direct fragment
    """
    v_mult = mdcg.vertices().attr(v_ind, MULT_ATTR)
    for i in range(0, 2 * (v_mult // 2), 2):
        yield (
            (v_ind, FORWARD_INT, i),
            (v_ind, FORWARD_INT, i + 1),
        )
        yield (
            (v_ind, REVERSE_INT, i),
            (v_ind, REVERSE_INT, i + 1),
        )


def dirf_builder(v: OccOrCT) -> tuple[DirFT, int]:
    """Direct fragments builder from vertex.

    Parameters
    ----------
    v : OccOrCT
        Multiplied doubled contig

    Returns
    -------
    DirFT
        Direct fragment associated to vertex
    int
        Position of vertex in built direct fragments
    """
    rest = v[COCC_IND] % 2
    return (
        (
            (v[CIND_IND], v[COR_IND], v[COCC_IND] - rest),
            (v[CIND_IND], v[COR_IND], v[COCC_IND] + (1 - rest)),
        ),
        rest,
    )


def dirf_canonical(v: OccOrCT) -> OccOrCT:
    """Direct fragment canonical.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig

    Returns
    -------
    OccOrCT
        Canonical of the direct fragments
    """
    return (v[CIND_IND], v[COR_IND], v[COCC_IND] - v[COCC_IND] % 2)


def dirf_other(v: OccOrCT) -> OccOrCT:
    """Return the other vertex in the direct fragments.

    Parameters
    ----------
    v : OccOrCT
        Multiplied oriented contig

    Returns
    -------
    OccOrCT
        The other vertex in the direct fragments
    """
    if v[COCC_IND] % 2 == 1:
        return (v[CIND_IND], v[COR_IND], v[COCC_IND] - 1)
    return (v[CIND_IND], v[COR_IND], v[COCC_IND] + 1)


# ---------------------------------------------------------------------------- #
#                                     PDirF                                    #
# ---------------------------------------------------------------------------- #
def pdirf(mdcg: MDCGraph) -> Iterator[PDirFT]:
    """Iterate on pair of direct fragments.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph

    Yields
    ------
    PInvFT
        Pair of direct fragments
    """
    repeated_contigs = mdcg.repeated_contigs()
    for i, v_ind in enumerate(repeated_contigs):
        for p in v_dirf(mdcg, v_ind):
            # same vertex index
            for q in v_dirf(mdcg, v_ind):
                if p[REVERSE_INT][COCC_IND] < q[FORWARD_INT][COCC_IND]:
                    yield p, q
            # superior vertex index
            for j in range(i + 1, len(repeated_contigs)):
                for q in v_dirf(mdcg, repeated_contigs[j]):
                    yield p, q


# ---------------------------------------------------------------------------- #
#                                     ADirF                                    #
# ---------------------------------------------------------------------------- #
def adirf(mdcg: MDCGraph) -> Iterator[EOccOrCT]:
    """Iterate on direct fragments successors.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contigs graph

    Yields
    ------
    EOccOrCT
        Canonical edges between two direct fragments
    """
    for u_ind in mdcg.repeated_contigs():
        u_mult = mdcg.vertices().attr(u_ind, MULT_ATTR)
        yield from __adirf_succs_f(mdcg, u_ind, u_mult)
        yield from __adirf_succs_r(mdcg, u_ind, u_mult)


def __adirf_succs_f(mdcg: MDCGraph,
                    u_ind: IndexT, u_mult: MultT) -> Iterator[EOccOrCT]:
    """Iterate of successors of forward vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contigs graph
    u_ind : IndexT
        Vertex identifier
    u_mult : MultT
        Vertex multiplicity

    Yields
    ------
    EOccOrCT
        Canonical edges between two direct fragments
    """
    # DOCU explain this
    for (v_ind, v_or), _ in mdcg.edges().succs((u_ind, FORWARD_INT)):
        #
        # Accept u_f -> v_f and u_f -> v_r
        # Will accept v_f -> u_f and v_f -> u_r
        #
        if u_ind != v_ind:
            v_mult = mdcg.vertices().attr(v_ind, MULT_ATTR)
            for u_occ in range(0, 2 * (u_mult // 2), 2):
                for v_occ in range(0, 2 * (v_mult // 2), 2):
                    yield (
                        (u_ind, FORWARD_INT, u_occ),
                        (v_ind, v_or, v_occ),
                    )
        #
        # Accept u_f -> u_f and u_f -> u_r
        #
        elif u_ind == v_ind:
            for u_occ in range(0, 2 * (u_mult // 2) - 2, 2):
                for v_occ in range(u_occ + 2, 2 * (u_mult // 2), 2):
                    yield (
                        (u_ind, FORWARD_INT, u_occ),
                        (u_ind, v_or, v_occ),
                    )


def __adirf_succs_r(mdcg: MDCGraph,
                    u_ind: IndexT, u_mult: MultT) -> Iterator[EOccOrCT]:
    """Iterate of successors of reverse vertex.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contigs graph
    u_ind : IndexT
        Vertex identifier
    u_mult : MultT
        Vertex multiplicity

    Yields
    ------
    EOccOrCT
        Canonical edges between two direct fragments
    """
    # DOCU explain this
    for (w_ind, w_or), _ in mdcg.edges().succs((u_ind, REVERSE_INT)):
        #
        # Accept u_r -> w_f and u_r -> w_r
        # Will accept w_r -> u_f and w_r -> u_r
        #
        if u_ind != w_ind:
            w_mult = mdcg.vertices().attr(w_ind, MULT_ATTR)
            for u_occ in range(0, 2 * (u_mult // 2), 2):
                for w_occ in range(0, 2 * (w_mult // 2), 2):
                    yield (
                        (u_ind, REVERSE_INT, u_occ),
                        (w_ind, w_or, w_occ),
                    )
        #
        # Accept u_r -> u_f
        #
        elif u_ind == w_ind and w_or == FORWARD_INT:
            for u_occ in range(0, 2 * (u_mult // 2) - 2, 2):
                for w_occ in range(u_occ + 2, 2 * (u_mult // 2), 2):
                    yield (
                        (u_ind, REVERSE_INT, u_occ),
                        (u_ind, FORWARD_INT, w_occ),
                    )


def adirf_other(edge: EOccOrCT) -> EOccOrCT:
    """Get the pairing edge.

    Parameters
    ----------
    edge : EOccOrCT
        Edge between two direct fragments

    Returns
    -------
    EOccOrCT
        The other edge between the two direct fragments
    """
    return dirf_other(edge[0]), dirf_other(edge[1])


# ---------------------------------------------------------------------------- #
#                               Consecutive PDirF                              #
# ---------------------------------------------------------------------------- #
def consecutive_pdirf(mdcg: MDCGraph, v_ind: IndexT) -> Iterator[PDirFT]:
    """Iterate over consecutive pairs of direct fragments.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph
    v_ind : IndexT
        Vertex' index

    Yields
    ------
    PDirFT
        Consecutive pairs of direct fragments
    """
    mult_v = mdcg.vertices().attr(v_ind, MULT_ATTR)
    for (i, j) in v_dirf(mdcg, v_ind):
        if j[COCC_IND] + 2 < mult_v:
            yield (
                (i, j),
                (
                    (i[CIND_IND], i[COR_IND], i[COCC_IND] + 2),
                    (j[CIND_IND], j[COR_IND], j[COCC_IND] + 2),
                ),
            )


# ============================================================================ #
#                               FORBIDDEN PAIRING                              #
# ============================================================================ #
def pdirf_alpha(mdcg: MDCGraph) -> Iterator[EOccOrCT]:
    """Iterate on alpha keys for direct fragments pairs.

    Parameters
    ----------
    mdcg : MDCGraph
        Multiplied doubled contig graph

    Yields
    ------
    EOccOrCT
        Alpha key set's element
    """
    # OPTIMIZE reduce loop redundancies in keys
    for (i, j), (k, l) in pdirf(mdcg):
        yield i, j  # . i j . or . j i .
        yield k, l  # . k l . or . l k .
        yield i, k  # i k l j or k i j l or l j i k or j l k i
        yield j, l
        yield i, l  # i l k j or l i j k or k j i l or j k l i
        yield j, k


def forbidden_nested(p_dirf: PDirFT) -> Iterator[ForbiddenPairDirFT]:
    """Iterate over the forbidden cases.

    Parameters
    ----------
    p_dirf : PDirFT
        Pair of direct fragments

    Yields
    ------
    ForbiddenPairDirFT
        Forbidden pairing cases for direct fragments,
        enriched with an index
    """
    (i, j), (k, l) = p_dirf
    # Forbidden order    # Alpha keys
    # ----------------------------------
    yield 0, i, k, l, j  # i k  k l  j l
    yield 1, i, l, k, j  # i l  k l  j k
    yield 2, j, k, l, i  # j k  k l  i l
    yield 3, j, l, k, i  # j l  k l  i k
    yield 4, k, i, j, l  # i k  i j  j l
    yield 5, k, j, i, l  # j k  i j  i l
    yield 6, l, i, j, k  # i l  i j  j k
    yield 7, l, j, i, k  # j l  i j  i k



# -*- coding=utf-8 -*-

"""Module for the RunMetadata class."""


from __future__ import annotations

from collections.abc import Iterable
from pathlib import Path
from typing import Any, Optional

import yaml  # type: ignore
from pulp import LpStatus, LpStatusOptimal
from revsymg.index_lib import IndexT

from khloraascaf.cli import (
    ARG_CONTIG_ATTRS,
    ARG_CONTIG_LINKS,
    ARG_CONTIG_STARTER,
    OPT_INSTANCE_NAME,
    OPT_MULT_UPB,
    OPT_OUTDEBUG,
    OPT_OUTDIR,
    OPT_PRESSCORE_UPB,
    OPT_SOLVER,
)
from khloraascaf.inputs import STR_ORIENT, IdCT, MultT, PresScoreT
from khloraascaf.lib import DR_REGION_ID, IR_REGION_ID, RegionIDT
from khloraascaf.multiplied_doubled_contig_graph import (
    CIND_IND,
    COCC_IND,
    COR_IND,
    OccOrCT,
)
from khloraascaf.outputs import (
    CONTIGS_OF_REGIONS_PREFIX,
    MAP_OF_REGIONS_PREFIX,
    ORIENT_INT_STR,
)
from khloraascaf.utils_debug import (
    FOUND_REPFRAG_PREFIX,
    VERTICES_OF_REGIONS_PREFIX,
)


# DOCU tuto metadata
# TOTEST metadata
# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
# DOCU IOConfig
# pylint: disable=too-many-instance-attributes
class IOConfig():
    """Input/Output config metadata class."""

    YAML_FILE: str = 'io_config.yaml'
    """IO config metadata file name."""

    @classmethod
    def from_run_directory(cls, ioconfig_dir: Path) -> IOConfig:
        """Intitialise IOConfig from the directory of the YAML.

        Parameters
        ----------
        ioconfig_dir : Path
            IO config metadata YAML directory

        Returns
        -------
        IOConfig
            IO config metadata
        """
        with open(ioconfig_dir / cls.YAML_FILE,
                  'r', encoding='utf-8') as f_in:
            io_dict: dict[str, Any] = yaml.load(f_in, yaml.Loader)
        return cls(
            ioconfig_dir,
            Path(io_dict[ARG_CONTIG_ATTRS]),
            Path(io_dict[ARG_CONTIG_LINKS]),
            io_dict[ARG_CONTIG_STARTER],
            MultT(io_dict[OPT_MULT_UPB]),
            PresScoreT(io_dict[OPT_PRESSCORE_UPB]),
            io_dict[OPT_SOLVER],
            Path(io_dict[OPT_OUTDIR]),
            io_dict[OPT_INSTANCE_NAME],
            bool(io_dict[OPT_OUTDEBUG]),
        )

    # pylint: disable=too-many-arguments
    def __init__(self, ioconfig_dir: Path, contig_attrs: Path,
                 contig_links: Path, contig_starter: IdCT,
                 mult_upb: MultT, presscore_upb: PresScoreT,
                 solver: str, outdir: Path, instance_name: str,
                 debug: bool):
        """The Initialiser.

        Parameters
        ----------
        ioconfig_dir : Path
            YAML file output directory path
        contig_attrs : Path
            Contig attributes file path
        contig_links : Path
            Contig links file path
        contig_starter : IdCT
            Starter contig's identifier
        mult_upb : MultT
            Multiplicities upper bound
        presscore_upb : PresScoreT
            Presence score upper bound
        solver : str
            MILP solver to use
        outdir : Path
            Output directory path
        instance_name : str
            Instance name
        debug : bool
            Output debug or not
        """
        self.__ioconfig_dir: Path = ioconfig_dir
        self.__contig_attrs: Path = contig_attrs
        self.__contig_links: Path = contig_links
        self.__contig_starter: IdCT = contig_starter
        self.__mult_upb: MultT = mult_upb
        self.__presscore_upb: PresScoreT = presscore_upb
        self.__solver: str = solver
        self.__outdir: Path = outdir
        self.__instance_name: str = instance_name
        self.__debug: bool = debug

    # ~*~ Getter ~*~

    def io_config_directory(self) -> Path:
        """IO config directory path.

        Returns
        -------
        Path
            IO config directory path
        """
        return self.__ioconfig_dir

    def contig_attrs(self) -> Path:
        """Contig attributes file path.

        Returns
        -------
        Path
            Contig attributes file path
        """
        return self.__contig_attrs

    def contig_links(self) -> Path:
        """Contig links file path.

        Returns
        -------
        Path
            Contig links file path
        """
        return self.__contig_links

    def contig_starter(self) -> IdCT:
        """Contig starter identifier.

        Returns
        -------
        IdCT
            Contig starter identifier
        """
        return self.__contig_starter

    def mult_upb(self) -> MultT:
        """Multiplicity upper-bound.

        Returns
        -------
        MultT
            Multiplicity upper-bound
        """
        return self.__mult_upb

    def presscore_upb(self) -> PresScoreT:
        """Presence score upper-bound.

        Returns
        -------
        PresScoreT
            Presence score upper-bound
        """
        return self.__presscore_upb

    def solver(self) -> str:
        """MILP solver identifier.

        Returns
        -------
        str
            MILP solver identifier
        """
        return self.__solver

    def outdir(self) -> Path:
        """Output directory given by the user.

        Returns
        -------
        Path
            Output directory.
        """
        return self.__outdir

    def instance_name(self) -> str:
        """Instance name.

        Returns
        -------
        str
            Instance name
        """
        return self.__instance_name

    def debug(self) -> bool:
        """Debug option value.

        Returns
        -------
        bool
            Debug option value
        """
        return self.__debug

    # ~*~ I/O ~*~

    def write_yaml(self):
        """Write to a YAML file.

        Warning
        -------
        Write a file.
        """
        with open(self.yaml_path(), 'w', encoding='utf-8') as yaml_out:
            yaml_out.write(
                yaml.dump(
                    {
                        ARG_CONTIG_ATTRS: str(self.__contig_attrs),
                        ARG_CONTIG_LINKS: str(self.__contig_links),
                        ARG_CONTIG_STARTER: self.__contig_starter,
                        OPT_MULT_UPB: self.__mult_upb,
                        OPT_PRESSCORE_UPB: self.__presscore_upb,
                        OPT_SOLVER: self.__solver,
                        OPT_OUTDIR: str(self.__outdir),
                        OPT_INSTANCE_NAME: self.__instance_name,
                        OPT_OUTDEBUG: self.__debug,
                    },
                    sort_keys=False,
                ),
            )

    def yaml_path(self) -> Path:
        """YAML file path.

        Returns
        -------
        Path
            YAML file path
        """
        return self.__ioconfig_dir / self.YAML_FILE


# DOCU MetadataSolution
class MetadataSolution():
    """Solution metadata class."""

    YAML_FILE: str = 'solution.yaml'
    """Solution metadata file name."""

    KEY_ILP_COMBINATION = 'ilp_combination'

    @classmethod
    def from_run_directory(cls, yaml_dir: Path) -> MetadataSolution:
        """Intitialise MetadataSolution from the YAML's directory.

        Parameters
        ----------
        yaml_dir : Path
            Solution metadata YAML directory

        Returns
        -------
        MetadataSolution
            Solution metadata
        """
        with open(yaml_dir / cls.YAML_FILE,
                  'r', encoding='utf-8') as f_in:
            sol_dict: dict[str, Any] = yaml.load(f_in, yaml.Loader)
        return cls(
            yaml_dir,
            sol_dict[cls.KEY_ILP_COMBINATION],
            Path(sol_dict[CONTIGS_OF_REGIONS_PREFIX]),
            Path(sol_dict[MAP_OF_REGIONS_PREFIX]),
        )

    # pylint: disable=too-many-arguments
    def __init__(self, solution_dir: Path,
                 ilp_combi: Iterable[RegionIDT],
                 contigs_of_regions_path: Path,
                 map_of_regions_path: Path):
        """The Initialiser.

        Parameters
        ----------
        solution_dir : Path
            Solution's path
        ilp_combi : iterable of RegionIDT
            Succession of region identifiers
        contigs_of_regions_path : Path
            Contigs of regions file path
        map_of_regions_path : Path
            Map of regions file path
        """
        self.__sol_dir: Path = solution_dir
        self.__ilp_combi: list[RegionIDT] = list(ilp_combi)
        self.__contigs_of_regions_path: Path = contigs_of_regions_path
        self.__map_of_regions_path: Path = map_of_regions_path

    # ~*~ Getter ~*~

    def solution_directory(self) -> Path:
        """Solution directory path.

        Returns
        -------
        Path
            Solution directory path
        """
        return self.__sol_dir

    def ilp_combination(self) -> list[RegionIDT]:
        """ILP succession.

        Returns
        -------
        list of RegionIDT
            ILP succession
        """
        return self.__ilp_combi

    def contigs_of_regions(self) -> Path:
        """Contigs of regions file path.

        Returns
        -------
        Path
            Contigs of regions file path
        """
        return self.__contigs_of_regions_path

    def map_of_regions(self) -> Path:
        """Map of regions file path.

        Returns
        -------
        Path
            Map of regions file path
        """
        return self.__map_of_regions_path

    # ~*~ I/O ~*~

    def write_yaml(self):
        """Write to a YAML file.

        Warning
        -------
        Write a file.
        """
        with open(self.yaml_path(), 'w', encoding='utf-8') as yaml_out:
            yaml_out.write(
                yaml.dump(
                    {
                        self.KEY_ILP_COMBINATION: self.__ilp_combi,
                        CONTIGS_OF_REGIONS_PREFIX:
                        str(self.__contigs_of_regions_path),
                        MAP_OF_REGIONS_PREFIX:
                        str(self.__map_of_regions_path),
                    },
                    sort_keys=False,
                ),
            )

    def yaml_path(self) -> Path:
        """YAML file path.

        Returns
        -------
        Path
            YAML file path
        """
        return self.__sol_dir / self.YAML_FILE


class MetadataAllSolutions():
    """Solutions metadata class."""

    YAML_FILE: str = 'solutions.yaml'
    """All solutions metadata file name."""

    KEY_ILP_COMBINATION = 'ilp_combination'

    @classmethod
    def from_run_directory(cls, yaml_dir: Path) -> MetadataAllSolutions:
        """Intitialise MetadataAllSolutions from the YAML's directory.

        Parameters
        ----------
        yaml_dir : Path
            Solution metadata YAML directory

        Returns
        -------
        MetadataAllSolutions
            Solutions metadata
        """
        with open(yaml_dir / cls.YAML_FILE,
                  'r', encoding='utf-8') as f_in:
            sol_list: list[dict[str, Any]] = yaml.load(f_in, yaml.Loader)
        return cls(
            yaml_dir,
            (
                MetadataSolution(
                    yaml_dir, sol_dict[cls.KEY_ILP_COMBINATION],
                    Path(sol_dict[CONTIGS_OF_REGIONS_PREFIX]),
                    Path(sol_dict[MAP_OF_REGIONS_PREFIX]),
                )
                for sol_dict in sol_list
            ),
        )

    # pylint: disable=too-many-arguments
    def __init__(self, solutions_dir: Path,
                 solutions: Iterable[MetadataSolution]):
        """The Initialiser.

        Parameters
        ----------
        solutions_dir : Path
            Solutions' directory path
        solutions : iterable of MetadataSolution
            Set of solution metadata
        """
        self.__sol_dir: Path = solutions_dir
        self.__solutions = list(solutions)

    # ~*~ Getter ~*~

    def solutions_directory(self) -> Path:
        """Solutions directory path.

        Returns
        -------
        Path
            Solutions directory path
        """
        return self.__sol_dir

    def solutions(self) -> list[MetadataSolution]:
        """List of solutions metadata.

        Returns
        -------
        list of MetadataSolution
            List of solutions metadata
        """
        return self.__solutions

    # ~*~ I/O ~*~

    def write_yaml(self, append_yaml: bool = False):
        """Write the solutions metadata to a YAML file.

        Parameters
        ----------
        append_yaml : bool
            If append to potentially already existing YAML file,
            by default False

        Warning
        -------
        Write a file.
        """
        write_mode = 'a' if append_yaml else 'w'

        with open(self.yaml_path(), write_mode, encoding='utf-8') as yaml_out:
            yaml_out.write(
                yaml.dump(
                    [
                        {
                            self.KEY_ILP_COMBINATION:
                            solution.ilp_combination(),
                            CONTIGS_OF_REGIONS_PREFIX:
                            str(solution.contigs_of_regions()),
                            MAP_OF_REGIONS_PREFIX:
                            str(solution.map_of_regions()),
                        } for solution in self.__solutions
                    ],
                    sort_keys=False,
                ),
            )

    def yaml_path(self) -> Path:
        """YAML file path.

        Returns
        -------
        Path
            YAML file path
        """
        return self.__sol_dir / self.YAML_FILE


# DOCU MetadataDebug
class MetadataDebug():
    """Debug metadata class."""

    YAML_FILE: str = 'debug.yaml'
    """Debug metadata file name."""

    KEY_ILP_COMBINATION = 'ilp_combination'
    KEY_STARTER_VERTEX = 'starter_vertex'
    KEY_OPT_VALUE = 'opt_value'
    KEY_ILP_STATUS = 'ilp_status'

    @classmethod
    def from_run_directory(cls, yaml_dir: Path) -> MetadataDebug:
        """Intitialise MetadataDebug from the YAML's directory.

        Parameters
        ----------
        yaml_dir : Path
            Debug metadata YAML directory

        Returns
        -------
        MetadataDebug
            Debug metadata
        """
        with open(yaml_dir / cls.YAML_FILE,
                  'r', encoding='utf-8') as f_in:
            debug_dict: dict[str, Any] = yaml.load(f_in, yaml.Loader)
        return cls(
            yaml_dir,
            debug_dict[cls.KEY_ILP_COMBINATION],
            str_to_vertex(debug_dict[cls.KEY_STARTER_VERTEX]),
            debug_dict[cls.KEY_ILP_STATUS],
            (
                float(debug_dict[cls.KEY_OPT_VALUE])
                if cls.KEY_OPT_VALUE in debug_dict else None
            ),
            (
                Path(debug_dict[VERTICES_OF_REGIONS_PREFIX])
                if VERTICES_OF_REGIONS_PREFIX in debug_dict else None
            ),
            (
                Path(debug_dict[MAP_OF_REGIONS_PREFIX])
                if MAP_OF_REGIONS_PREFIX in debug_dict else None
            ),
            (
                Path(debug_dict[FOUND_REPFRAG_PREFIX])
                if FOUND_REPFRAG_PREFIX in debug_dict else None
            ),
        )

    # pylint: disable=too-many-arguments
    def __init__(self, run_dir: Path,
                 ilp_combi: Iterable[RegionIDT],
                 starter_vertex: OccOrCT,
                 ilp_status: str,
                 opt_value: float | None = None,
                 vertices_of_regions_path: Path | None = None,
                 map_of_regions_path: Path | None = None,
                 repeat_fragments_path: Path | None = None):
        """The Initialiser.

        Parameters
        ----------
        run_dir : Path
            Run path
        ilp_combi : iterable of RegionIDT
            Succession of region identifiers
        starter_vertex : OccOrCT
            Starter vertex
        ilp_status : str
            ILP solver status
        opt_value : float | None
            Optimal ILP value, optional
        vertices_of_regions_path : Path | None
            Vertices of regions file path, optional
        map_of_regions_path : Path | None
            Map of regions file path, optional
        repeat_fragments_path : Path | None
            Repeated fragments path, optional
        """
        self.__run_dir: Path = run_dir
        self.__ilp_combi: list[RegionIDT] = list(ilp_combi)
        self.__starter_vertex: OccOrCT = starter_vertex
        self.__ilp_status: str = ilp_status
        self.__opt_value: float | None = opt_value
        self.__vertices_of_regions_path: Path | None = vertices_of_regions_path
        self.__map_of_regions_path: Path | None = map_of_regions_path
        self.__repeat_fragments_path: Path | None = repeat_fragments_path

    # ~*~ Getter ~*~

    def run_dir(self) -> Path:
        """Run directory path.

        Returns
        -------
        Path
            Run directory path
        """
        return self.__run_dir

    def ilp_combination(self) -> list[RegionIDT]:
        """ILP succession.

        Returns
        -------
        list of RegionIDT
            ILP succession
        """
        return self.__ilp_combi

    def starter_vertex(self) -> OccOrCT:
        """Starter vertex.

        Returns
        -------
        OccOrCT
            Starter vertex
        """
        return self.__starter_vertex

    def ilp_status(self) -> str:
        """ILP string status.

        Returns
        -------
        str
            ILP string status
        """
        return self.__ilp_status

    def opt_value(self) -> Optional[float]:
        """ILP optimal value if any.

        Returns
        -------
        float | None
            Optimal value, else None
        """
        return self.__opt_value

    def vertices_of_regions(self) -> Path:
        """Vertices of regions file path.

        Returns
        -------
        Path
            Vertices of regions file path

        Raises
        ------
        FileNotFoundError
            File does not exist
        """
        if self.__vertices_of_regions_path is None:
            raise FileNotFoundError()
        return self.__vertices_of_regions_path

    def map_of_regions(self) -> Path:
        """Map of regions file path.

        Returns
        -------
        Path
            Map of regions file path

        Raises
        ------
        FileNotFoundError
            File does not exist
        """
        if self.__map_of_regions_path is None:
            raise FileNotFoundError()
        return self.__map_of_regions_path

    def repeat_fragments(self) -> Path:
        """Repeated fragments file path.

        Returns
        -------
        Path
            Repeated fragments file path

        Raises
        ------
        FileNotFoundError
            File does not exist
        """
        if self.__repeat_fragments_path is None:
            raise FileNotFoundError()
        return self.__repeat_fragments_path

    # ~*~ I/O ~*~

    def write_yaml(self):
        """Write the debug metadata to a YAML file.

        Warning
        -------
        Write a file.
        """
        dict_debug: dict[str, Any] = {
            self.KEY_ILP_COMBINATION: self.__ilp_combi,
            self.KEY_STARTER_VERTEX:
            vertex_to_str(self.__starter_vertex),
            self.KEY_ILP_STATUS: self.__ilp_status,
        }
        if self.__ilp_status == LpStatus[LpStatusOptimal]:
            assert self.__opt_value is not None
            assert self.__vertices_of_regions_path is not None
            assert self.__map_of_regions_path is not None
            dict_debug.update(
                {
                    self.KEY_OPT_VALUE: self.__opt_value,
                    VERTICES_OF_REGIONS_PREFIX:
                    str(self.__vertices_of_regions_path),
                    MAP_OF_REGIONS_PREFIX:
                    str(self.__map_of_regions_path),
                },
            )
            if self.__ilp_combi[-1] in (IR_REGION_ID, DR_REGION_ID):
                assert self.__repeat_fragments_path is not None
                dict_debug.update(
                    {
                        FOUND_REPFRAG_PREFIX:
                        str(self.__repeat_fragments_path),
                    },
                )

        with open(self.yaml_path(), 'w', encoding='utf-8') as yaml_out:
            yaml_out.write(yaml.dump(dict_debug, sort_keys=False))

    def yaml_path(self) -> Path:
        """YAML file path.

        Returns
        -------
        Path
            YAML file path
        """
        return self.__run_dir / self.YAML_FILE


class MetadataAllDebugs():
    """Debugs metadata class."""

    YAML_FILE: str = 'debugs.yaml'
    """All debugs metadata file name."""

    KEY_ILP_COMBINATION = 'ilp_combination'
    KEY_STARTER_VERTEX = 'starter_vertex'
    KEY_OPT_VALUE = 'opt_value'
    KEY_ILP_STATUS = 'ilp_status'

    @classmethod
    def from_run_directory(cls, yaml_dir: Path) -> MetadataAllDebugs:
        """Intitialise MetadataAllDebugs from the YAML's directory.

        Parameters
        ----------
        yaml_dir : Path
            Debugs metadata YAML directory

        Returns
        -------
        MetadataAllDebugs
            Debugs metadata
        """
        with open(yaml_dir / cls.YAML_FILE,
                  'r', encoding='utf-8') as f_in:
            debug_list: list[dict[str, Any]] = yaml.load(f_in, yaml.Loader)
        return cls(
            yaml_dir,
            (
                MetadataDebug(
                    yaml_dir, debug_dict[cls.KEY_ILP_COMBINATION],
                    str_to_vertex(debug_dict[cls.KEY_STARTER_VERTEX]),
                    debug_dict[cls.KEY_ILP_STATUS],
                    (
                        float(debug_dict[cls.KEY_OPT_VALUE])
                        if cls.KEY_OPT_VALUE in debug_dict else None
                    ),
                    (
                        Path(debug_dict[VERTICES_OF_REGIONS_PREFIX])
                        if VERTICES_OF_REGIONS_PREFIX in debug_dict else None
                    ),
                    (
                        Path(debug_dict[MAP_OF_REGIONS_PREFIX])
                        if MAP_OF_REGIONS_PREFIX in debug_dict else None
                    ),
                    (
                        Path(debug_dict[FOUND_REPFRAG_PREFIX])
                        if FOUND_REPFRAG_PREFIX in debug_dict else None
                    ),
                )
                for debug_dict in debug_list
            ),
        )

    # pylint: disable=too-many-arguments
    def __init__(self, debugs_dir: Path,
                 debugs: Iterable[MetadataDebug]):
        """The Initialiser.

        Parameters
        ----------
        debugs_dir : Path
            Debugs' directory path
        debugs : iterable of MetadataDebug
            Set of debug metadata
        """
        self.__run_dir: Path = debugs_dir
        self.__debugs = list(debugs)

    # ~*~ Getter ~*~

    def runs_directory(self) -> Path:
        """Runs directory path.

        Returns
        -------
        Path
            Runs directory path
        """
        return self.__run_dir

    def debugs(self) -> list[MetadataDebug]:
        """List of debugs metadata.

        Returns
        -------
        list of MetadataDebug
            List of debugs metadata
        """
        return self.__debugs

    # ~*~ I/O ~*~

    def write_yaml(self, append_yaml: bool = False):
        """Write the debugs metadata to a YAML file.

        Parameters
        ----------
        append_yaml : bool
            If append to potentially already existing YAML file,
            by default False

        Warning
        -------
        Write a file.
        """
        debug_list: list = []
        for debug in self.__debugs:
            dict_debug: dict[str, Any] = {
                self.KEY_ILP_COMBINATION: debug.ilp_combination(),
                self.KEY_STARTER_VERTEX:
                vertex_to_str(debug.starter_vertex()),
                self.KEY_ILP_STATUS: debug.ilp_status(),
            }
            if debug.ilp_status() == LpStatus[LpStatusOptimal]:
                assert debug.opt_value() is not None
                assert debug.vertices_of_regions() is not None
                assert debug.map_of_regions() is not None
                dict_debug.update(
                    {
                        self.KEY_OPT_VALUE: debug.opt_value(),
                        VERTICES_OF_REGIONS_PREFIX:
                        str(debug.vertices_of_regions()),
                        MAP_OF_REGIONS_PREFIX:
                        str(debug.map_of_regions()),
                    },
                )
                if debug.ilp_combination()[-1] in (IR_REGION_ID, DR_REGION_ID):
                    assert debug.repeat_fragments() is not None
                    dict_debug.update(
                        {
                            FOUND_REPFRAG_PREFIX:
                            str(debug.repeat_fragments()),
                        },
                    )
            debug_list.append(dict_debug)

        write_mode = 'a' if append_yaml else 'w'
        with open(self.yaml_path(), write_mode, encoding='utf-8') as yaml_out:
            yaml_out.write(yaml.dump(debug_list, sort_keys=False))

    def yaml_path(self) -> Path:
        """YAML file path.

        Returns
        -------
        Path
            YAML file path
        """
        return self.__run_dir / self.YAML_FILE


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                               String Formatter                               #
# ---------------------------------------------------------------------------- #
def vertex_to_str(vertex: OccOrCT) -> str:
    """Transform a vertex to a string.

    Parameters
    ----------
    vertex : OccOrCT
        Vertex

    Returns
    -------
    str
        String representation
    """
    return (
        f'{vertex[CIND_IND]}'
        f'\t{ORIENT_INT_STR[vertex[COR_IND]]}'
        f'\t{vertex[COCC_IND]}'
    )


def str_to_vertex(vertex_str: str) -> OccOrCT:
    """Transform a vertex string to the vertex.

    Parameters
    ----------
    vertex_str : str
        Vertex string

    Returns
    -------
    OccOrCT
        Vertex
    """
    str_list = vertex_str.split('\t')
    return (
        IndexT(str_list[CIND_IND]),
        STR_ORIENT[str_list[COR_IND]],  # type: ignore
        IndexT(str_list[COCC_IND]),
    )
